/*
 * GluonJ, An Aspect-Oriented Programming System for Java.
 * Copyright (C) 2006 Shigeru Chiba. All Rights Reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License.  Alternatively, the contents of this file may be used under
 * the terms of the GNU Lesser General Public License Version 2.1 or later.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 */
package javassist.gluonj;

import javassist.ClassPool;
import javassist.gluonj.pc.*;
import javassist.gluonj.weave.AliasList;


public class Pointcut {
    private PointcutNode tree;
    private Pointcut prev;
    private AliasList aliases;
    public Pointcut and;    // AND is stronger than OR
    public Pointcut or;

    Pointcut(Pointcut back) {
        tree = null;
        prev = back;
        aliases = null;
        and = null;
        or = null;
    }

    public String toString() {
        AliasList a = new AliasList();
        String str = parse(a).toString();
        if (a.next == null)
            return str;
        else
            return a.next.toString() + " " + str; 
    }

    static class Iterator {
        Pointcut node;
    }

    PointcutNode parse(String glueName, ClassPool cp, AliasList a)
        throws WeaveException
    {
        PointcutNode pcn = parse(a);
        prepare(new Parser(glueName, cp), pcn);
        return pcn;
    }

    private static void prepare(Parser parser, PointcutNode pcn)
        throws WeaveException
    {
        pcn.accept(new PrepareVisitor(parser));
    }

    static class PrepareVisitor implements PointcutVisitor {
        private Parser parser;

        public PrepareVisitor(Parser p) {
            parser = p;
        }

        public void visit(AndPc n) throws WeaveException {
            n.first().accept(this);
            AndPc r = n.rest();
            if (r != null)
                r.accept(this);
        }

        public void visit(OrPc n) throws WeaveException {
            n.first().accept(this);
            OrPc r = n.rest();
            if (r != null)
                r.accept(this);
        }

        public void visit(WithinPc pc) throws WeaveException {
            pc.prepare(parser);
        }

        public void visit(CallPc pc) throws WeaveException {
            pc.prepare(parser);
        }

        public void visit(GetPc pc) throws WeaveException {
            pc.prepare(parser);
        }

        public void visit(SetPc pc) throws WeaveException {
            pc.prepare(parser);
        }

        public void visit(IfPc pc) throws WeaveException {
            pc.prepare(parser);
        }

        public void visit(CflowPc pc) throws WeaveException {
            pc.prepare(parser);
        }
    }

    private PointcutNode parse(AliasList allAliases) {
        Iterator it = new Iterator();
        it.node = this;
        OrPc expr = parseOr(it, null, allAliases);
        if (expr.rest() == null)
            return expr.first();
        else
            return expr;
    }

    OrPc parseOr(Iterator it, OrPc tail, AliasList allAliases) {
        Pointcut pc;
        do {
            AndPc left = parseAnd(it, null, allAliases);
            tail = new OrPc(left.rest() == null ? left.first() : left, tail);
            if ((pc = it.node) == null)
                break;

            it.node = pc.prev;
        } while (it.node != null && it.node.or == pc);
        return tail;
    }

    AndPc parseAnd(Iterator it, AndPc tail, AliasList allAliases) {
        Pointcut pc;
        do {
            if ((pc = it.node) == null)
                break;

            tail = new AndPc(pc.tree, tail);
            if (pc.aliases != null)
                allAliases.next = AliasList.concat(allAliases.next, pc.aliases);

            it.node = pc.prev;
        } while (it.node != null && it.node.and == pc);
        it.node = pc;
        return tail;
    }

    Pointcut setAlias(AliasList a) {
        aliases = a;
        return this;
    }

    private void setTree(PointcutNode node) {
        if (tree != null)
            throw new RuntimeException("syntax error in pointcut");

        tree = node;
        and = new Pointcut(this);
        or = new Pointcut(this);
    }

    public Pointcut expr(Pointcut pc) {
        AliasList a = new AliasList();
        setTree(pc.parse(a));
        aliases = a.next;
        return this;
    }

    public Pointcut within(String className) {
        setTree(new WithinPc(className));
        return this;
    }

    public Pointcut call(String method) {
        setTree(new CallPc(method));
        return this;
    }

    public Pointcut set(String field) {
        setTree(new SetPc(field));
        return this;
    }

    public Pointcut get(String field) {
        setTree(new GetPc(field));
        return this;
    }

    /**
     * AspectJ's if pointcut designator.
     */
    public Pointcut when(String expr) {
        setTree(new IfPc(expr));
        return this;
    }

    public Pointcut cflow(String method) {
        setTree(new CflowPc(method));
        return this;
    }
}
