/*
 * GluonJ, An Aspect-Oriented Programming System for Java.
 * Copyright (C) 2006 Shigeru Chiba. All Rights Reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License.  Alternatively, the contents of this file may be used under
 * the terms of the GNU Lesser General Public License Version 2.1 or later.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 */
package javassist.gluonj.plugin;

import javassist.NotFoundException;
import javassist.gluonj.*;
import javassist.gluonj.pc.*;
import javassist.expr.MethodCall;
import javassist.expr.FieldAccess;
import javassist.bytecode.MethodInfo;
import javassist.bytecode.FieldInfo;
import javassist.bytecode.AnnotationsAttribute;
import javassist.bytecode.annotation.Annotation;

@Glue public class MetaTag {

    /* Comment out the next line if you don't want to see a stack trace.
     */
    // @Refine("javassist.gluonj.Gluon")
    public static class Debug {
        // print a stack trace when an error message is printed.
        public static boolean stackTrace = true;
    }

    public interface AnnotatePcVisitor {
        void visit(AnnotatePc pc) throws WeaveException;
    }

    public static class AnnotatePc extends PointcutNode {
        private String tag;
        private PcPattern pattern;

        public AnnotatePc(String tag) {
            this.tag = tag;
        }

        public String toString() {
            return "annotate(" + tag + ")";
        }

        public void prepare(Parser p) throws WeaveException {
            pattern = p.parseClass(removeAt(tag));
            if (!pattern.isClassName())
                throw new WeaveException("bad pattern: " + toString());
        }

        private static String removeAt(String name) {
            if (name != null && name.length() > 1 && name.charAt(0) == '@')
                return name.substring(1);
            else
                return name;
        }

        public void accept(PointcutVisitor v) throws WeaveException {
            ((AnnotatePcVisitor)v).visit(this);
        }

        public boolean match(AnnotationsAttribute attr) throws WeaveException {
            if (attr == null)
                return false;

            Annotation[] anno = attr.getAnnotations();
            int n = anno.length;
            for (int i = 0; i < n; i++)
                if (pattern.matchClass(anno[i].getTypeName()))
                    return true;

            return false;
        }
    }

    @Refine("javassist.gluonj.pc.PointcutVisitor")
    interface PcVisitor extends AnnotatePcVisitor {}

    @Refine("javassist.gluonj.Pointcut.PrepareVisitor")
    public static class Prepare {
        Parser parser;
        public void visit(AnnotatePc pc) throws WeaveException {
            pc.prepare(parser);
        }
    }

    @Refine("javassist.gluonj.Gluon.Matcher")
    public static class Match {
        protected boolean result;
        protected Residue residue;

        public void visit(AnnotatePc pc) throws WeaveException {
            result = false;
        }
    }

    @Refine("javassist.gluonj.Gluon.CallMatcher")
    public static class CallMatch {
        private MethodCall joinPoint;
        protected boolean result;

        public void visit(AnnotatePc pc) throws WeaveException {
            try {
                MethodInfo minfo = joinPoint.getMethod().getMethodInfo2();
                AnnotationsAttribute aa1 = (AnnotationsAttribute)
                    minfo.getAttribute(AnnotationsAttribute.invisibleTag);
                AnnotationsAttribute aa2 = (AnnotationsAttribute)
                    minfo.getAttribute(AnnotationsAttribute.visibleTag);
                result = pc.match(aa1) || pc.match(aa2);
            }
            catch (NotFoundException e) {
                throw new WeaveException(e);
            }
        }
    }

    @Refine("javassist.gluonj.Gluon.FieldMatcher")
    public static class FieldMatch {
        boolean result;
        private FieldAccess joinPoint;

        public void visit(AnnotatePc pc) throws WeaveException {
            try {
                FieldInfo finfo = joinPoint.getField().getFieldInfo2();
                AnnotationsAttribute aa1 = (AnnotationsAttribute)
                    finfo.getAttribute(AnnotationsAttribute.invisibleTag);
                AnnotationsAttribute aa2 = (AnnotationsAttribute)
                    finfo.getAttribute(AnnotationsAttribute.visibleTag);
                result = pc.match(aa1) || pc.match(aa2);
            }
            catch (NotFoundException e) {
                throw new WeaveException(e);
            }
        }
    }

    @Refine("javassist.gluonj.Gluon.CflowCollector")
    public static class Cflow {
        public void visit(AnnotatePc pc) throws WeaveException {
            // do nothing.
        }
    }

    public interface AnnotateAvailable {
        Pointcut annotate(String tag);
    }

    @Refine("javassist.gluonj.Pointcut")
    static abstract class Pcut implements AnnotateAvailable {
        abstract void setTree(PointcutNode pcn);

        public Pointcut annotate(String tag) {
            setTree(new AnnotatePc(tag));
            return (Pointcut)Gluon.$refine(this);
        }
    }

    @Refine("javassist.gluonj.Pcd")
    static class Pcd {
        @Abstract private static Pointcut make() { return null; }

        public static Pointcut annotate(String tag) {
            AnnotateAvailable pc = (AnnotateAvailable)make();
            return pc.annotate(tag);
        }
    }
}
