/*
 * GluonJ, An Aspect-Oriented Programming System for Java.
 * Copyright (C) 2006 Shigeru Chiba. All Rights Reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License.  Alternatively, the contents of this file may be used under
 * the terms of the GNU Lesser General Public License Version 2.1 or later.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 */
package javassist.gluonj.weave;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.Set;

import javassist.CtClass;
import javassist.CannotCompileException;
import javassist.NotFoundException;
import javassist.gluonj.Gluon;
import javassist.gluonj.WeaveException;

/**
 * Another load-time weaver.  Provided only for testing purpose.
 */
public class RuntimeWeaver extends Weaver {
    private ClassLoader classLoader;
    private boolean running;

    /**
     * Constructs a weaver.
     *
     * @param g                 glue object.
     * @param mainClass         the woven classes are loaded by the same
     *                          class loader as this <code>mainClass</code>.  
     */
    public RuntimeWeaver(Gluon g, Class mainClass) {
        super(g);
        classLoader = mainClass.getClassLoader();
        running = false;
    }

    /**
     * Executes weaving.  This transforms only the classes that refine
     * classes are applied to or the super classes of those classes.
     */
    public void run() throws RuntimeException {
        running = true;
        try {
            Set set = refineTargets.entrySet();
            Iterator it = set.iterator();
            while (it.hasNext()) {
                Map.Entry e =(Map.Entry)it.next();
                String targetName = (String)e.getKey();
                LinkedList refines = (LinkedList)e.getValue();
                transform(targetName, refines);
            }
        }
        catch (WeaveException e) {
            throw new RuntimeException(e.getMessage(), e);
        }
    }

    /**
     * Weaves advices and a given class.
     *
     * @param className         the fully-qualified name of the class.
     */
    public void use(String className) throws RuntimeException {
        if (!running)
            run();

        try {
            transform(className);
        }
        catch (WeaveException err) {
            throw new RuntimeException(err.getMessage(), err);
        }
    }

    protected boolean transform(String className, LinkedList refines)
        throws WeaveException
    {
        if (className.startsWith("java.") || className.startsWith("javax."))
            return false;     // cannot process basic classes.

        try {
            CtClass target = pool.get(className);
            if (target.isFrozen())
                return false;

            CtClass superClass = target.getSuperclass();
            if (superClass != null)
                transform(superClass.getName());

            return super.transform(className, refines);
        }
        catch (NotFoundException nfe) {
            throw new WeaveException(nfe);
        }
    }

    protected void save(CtClass clazz, boolean modify) throws CannotCompileException {
        try {
            clazz.toClass(classLoader);
        }
        catch (CannotCompileException e) {
            Throwable cause = e.getCause();
            if (cause != null && (cause instanceof LinkageError))
                throw (LinkageError)cause;
            else
                throw e;
        }
    }
}
